//
// Cyrano Chess engine
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//

#include "magicmoves.h"
#include "engine.hpp"
#include "gproto.hpp"
#include "hash.hpp"
#include "board.hpp"
#include "moves.hpp"
#include "stats.hpp"
#include "search.hpp"
#include "eval.hpp"
#include "genmagic.hpp"
#include <stdlib.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

// enables the PVS search, undefine to have standard alpha beta search (for bug hunting)
// reduce the number of nodes
#define NULL_WINDOW

// reduce the number of nodes
#define NULL_MOVE
//#define TWO_NULL

// allways extend depth by one ply when in check
// increase the number of nodes, find mates in less plies
#define EXTEND_CHECK
// extend the pv a bit, the score should be more reliable, and this gives a better move ordering for the next iteration
//#define SOME_PV_EXTENSION
//#define NM_THREAT_DETECTION

#define MINIMIZE

//    #define FUTILITY_PRUNING
    #define RAZORING

//    #define BM_EXTENSION
    // research in reduced disbaled too
#define LESS_CHECK

//#define QS_VERIFY_EXTENSION

//#define QS_CHECK_LIMIT  (pste->plyRem > 0)
//#define QS_CHECK_LIMIT  (pste->plyRem >= 0)
#define QS_CHECK_LIMIT  (pste->plyRem >= 0)
#define QS_CAPT_LIMIT  (pste->plyRem >= 0)
//#define QS_CAPT_LIMIT  (false)
//#define QS_CHECK_LIMIT  false

// reduce the number of nodes
#define USE_HASH
// change move ordering => reduce the number of nodes
#define USE_KILLERS

#define QS_BEFORE_HASH      // note:probing hash first gives worse results

// late move reduction
#define USE_LMR // disabled to test things
// skip descending the tree for some useless xchange in the quiescent search
// reduce the number of nodes ?
#define PRUNE_QS
// ...
// this will stop search when we have a score good enought (mate)
#define DISTANCE_PRUNING

#define RECAPTURE_EXTENSION

#define MOVE_SAFETY_SORT  // this is not so bad
//#define MOVE_PST_SORT     // this is not so good
#define MOVE_HISTORY_SORT

#define USE_IID
//#define HASH_NM       // this is a problem in endgame ???
#define EVAL_INTERIOR
#define USE_RECOGNIZER

#define moveisCapture(pcm) (Capt((pcm)->m) != 0)
#define moveisEqualCapture(pcm) (((pcm)->cmk & (cmkEQUCAPT)) != 0)
#define moveisGoodCapture(pcm) (((pcm)->cmk & (cmkGOODCAPT|cmkEQUCAPT)) != 0)
#define moveisVeryGoodCapture(pcm) (((pcm)->cmk & (cmkGOODCAPT)) != 0)
#define moveisBadCapture(pcm) (((pcm)->cmk & (cmkBADCAPT)) != 0)
#define moveisPromotion(pcm) ((Pc((pcm)->m)|1) == (B_PROM|1))
#define moveisPawnMove(pcm) ((Pc((pcm)->m)|1) == (PAWN|1))
#define moveto7thRank(pcm)  ((Rank(To((pcm)->m)) == 6) || (Rank(To((pcm)->m)) == 1))
//#define moveisGood(pcm) (((pcm)->cmk & (cmkMASK)) >= cmkQUEEN)
#define moveisGood(pcm) (((pcm)->cmk & (cmkMASK)) >= cmkEQUCAPT)



#define FRONTIER    0
#define isFrontier(pste)    ((pste)->plyRem == FRONTIER)
#define isRazorDepth(pste)    ((pste)->plyRem == (FRONTIER+2))

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta, bool do_check);
static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, nodeType nt);
static int SearchPV(PCON pcon, PSTE pste, int Alpha, int Beta);
static int INLINE capture_value(PCON pcon, const STE *pste, const CM *pcm);

//#define BM_DEPTH    2
#define BM_DEPTH    3
#define MAXEXT(n)   (50)
//#define MAXEXT(n)   (20+(n))
//#define MAXEXT(n)   (12+(n))
//#define MAXEXT(n)   (12)
//#define MAXEXT(n)   (70)
// 1 => 20
// 10 => 30
// 20 => 35
// 20 + n/2     1=>20     5=>22     10=>25    15=>27    20=>30  25=>32  30=>35  (40=>40)
//#define MAXEXT(n)   (20+(n)/2)
// 10 + n/2     1=>10=>11     5=>12=>17     10=>15=>25    15=>17=>32    20=>20  25=>22  30=>25  (20=>20)
//#define MAXEXT(n)   (1)


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static void SortKillers(PSTE pste, int depth)
{
    U32 Killer1 = Killers[depth].Killer1.m;
    U32 Killer2 = Killers[depth].Killer2.m;
    U32 Killer3 = Killers[depth-2].Killer1.m;
    U32 Killer4 = Killers[depth-2].Killer2.m;
    foreachmove(pste, pcm)
        if( (pcm->m == MateKillers[depth].Killer1.m) || (pcm->m == MateKillers[depth].Killer2.m)) {
			pcm->cmk |= cmkKILLER | 5;
            stats.Killers++;
        } else if ( (pcm->m == Killer1) || (pcm->m == Killer2) ) {
			pcm->cmk |= cmkKILLER | 3;
            stats.Killers++;
		}/* else if ( (pcm->m == Killer3) || (pcm->m == Killer4) ) {
			pcm->cmk |= cmkKILLER | 1;
            stats.Killers++;
		}*/
}
static INLINE int MaterialGain(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
    return b_valPc[ Capt(pcm->m) ];
}


//	This will be used to figure out which move to search first.  It is very
//	important to try to search the best move first, since this will produce
//	a huge reduction in tree size.

//	Best moves are typically winning captures and moves that have been shown
//	to be good here in the past (PV move and hash table move).

static void PrepareSort(CON *pcon, const STE *pste, int depth, PHASHPawn phashp) {
    // bitmap of squares attacked by the opponent pawns
    bitboard notSafe = phashp->side[pste->side ^1 ].pawnatk;
//    int dRank = 7;
//    if( pste->side == B )
//        dRank = 0;

    foreachmove(pste, pcm) {
        pcm->cmk = 0;
        if(Capt(pcm->m)) {
            pcm->cmk = MVV_LVA[Pc(pcm->m)][Capt(pcm->m)];
            if( pcm->cmk & cmkBADCAPT ) {
                // RxN, QxP
                int val = capture_value(pcon, pste, pcm );
                if( val > 0)
                    //pcm->cmk = cmkGOODCAPT | (2000+val);
                    //pcm->cmk = cmkGOODCAPT | (pcm->cmk & ~cmkMASK);
                    pcm->cmk = cmkGOODCAPT;
                else if( val == 0 )
                    pcm->cmk = cmkEQUCAPT | 2000;
                //else 
                //    pcm->cmk = cmkBADCAPT | (2000+val);
            }
        }
        else if( moveisPawnMove(pcm) && moveto7thRank(pcm)) {
            // only checking the dest rank because we know it's a pawn
//            pcm->cmk = (cmkKILLER | cmkGOODCAPT) + 10000;
            pcm->cmk = cmkQUEEN;
        } else if( moveisPromotion(pcm) ) {
            if( Prom(pcm->m) == PROM_Q )
                pcm->cmk = cmkGOODCAPT | cmkQUEEN;
            else
                pcm->cmk = cmkQUEEN | (100 - Prom(pcm->m));
        } else {
            unsigned int history_score = 0;

#ifdef MOVE_HISTORY_SORT
            history_score = History.sort_key( pcm );
            pcm->cmk = cmkNONE + history_score;
#else
            pcm->cmk = cmkNONE + 1000;
            if( AnyColor(Pc(pcm->m)) == AnyColor(B_CASTLE) )
                pcm->cmk = cmkQUEEN;    // will not be reduced

#ifdef MOVE_SAFETY_SORT
            {
                if(!moveisPawnMove(pcm)) {
                    bitboard to = IdxToU64(To(pcm->m));
                    // moving to a square attacked by a pawn
                    if( notSafe & to )
                        pcm->cmk -= 100;
                    bitboard from = IdxToU64(From(pcm->m));
                    // leaving a square attacked by a pawn
                    if( notSafe & from )
                        pcm->cmk += 50;
                }
            }
#endif
            {
//                pcm->cmk += (dRank - Rank(From(pcm->m))) - (dRank - Rank(To(pcm->m)));
                int d = Rank(To(pcm->m)) - (Rank(From(pcm->m))>>1);
                if( pste->side == B )
                    d = 7-d;
                if( d > 0 && AnyColor(Pc(pcm->m)) == AnyColor(KNIGHT) )
                    d += 10;
                pcm->cmk += d;
            }
#endif

#ifdef MOVE_PST_SORT
        if( history_score == 0 ) {
            // set cmk to delta positional value
            int pc = Pc(pcm->m);
            int neutralPc = pc & (~1);
            int co = pc & 1;
            // neutral color square for positional values
            int nSqr = remapSq[co][To(pcm->m)];
            // basic positional value
            pcm->cmk += staticPosVal[neutralPc >> 1][nSqr];
            nSqr = remapSq[co][From(pcm->m)];
            // basic positional value
            pcm->cmk -= staticPosVal[neutralPc >> 1][nSqr];
        }
#endif
        }

    }
}

static void PrepareSortQS(PCON pcon, const STE *pste, int depth) {
    foreachmove(pste, pcm) {
        U32 m = pcm->m;
        pcm->cmk = 0;
        if( moveisPromotion(pcm) ) {
            if( Prom(pcm->m) == PROM_Q )
                pcm->cmk = cmkGOODCAPT | cmkQUEEN;
            else
                pcm->cmk = cmkQUEEN | (100 - Prom(pcm->m));
        } else if(Capt(m)) {
            pcm->cmk = MVV_LVA[Pc(m)][Capt(m)];
            if( pcm->cmk & cmkBADCAPT ) {
                // RxN, QxP
                int val = capture_value(pcon, pste, pcm );
                if( val > 0)
                    //pcm->cmk = cmkGOODCAPT | (2000+val);
                    //pcm->cmk = cmkGOODCAPT | (pcm->cmk & ~cmkMASK);
                    pcm->cmk = cmkGOODCAPT;
                else if( val == 0 )
                    pcm->cmk = cmkEQUCAPT | 2000;
                //else 
                //    pcm->cmk = cmkBADCAPT | (2000+val);
            }
        } else if( moveisPawnMove(pcm) && moveto7thRank(pcm)) {
            if( see(pcon, pste, From(pcm->m), To(pcm->m)) >= 0 )
                pcm->cmk = cmkEQUCAPT | 0;
        }
    }
}

static INLINE void SortNext(const STE *pste, PCM pcm) {
	PCM	pcmBest = pcm;

	for (PCM pcmI = pcm + 1; pcmI < (pste + 1)->mlist_start ; pcmI++)
		if (pcmI->cmk > pcmBest->cmk)
			pcmBest = pcmI;
	CM cm = *pcm;
	*pcm = *pcmBest;
	*pcmBest = cm;
}

static INLINE void copyPV(PSTE pste, const CM &cm) {
    int how_many = (pste + 1)->ccmPv;
    pste->argcmPv[0] = cm;
    pste->ccmPv = how_many + 1;
    // the loop can copy one more
    for(int i = 0; i < how_many; i += 2) {
        pste->argcmPv[i+1] = (pste+1)->argcmPv[i];
        pste->argcmPv[i+2] = (pste+1)->argcmPv[i+1];
    }
}
static INLINE void copyPVQS(PSTE pste, const CM &cm) {
    copyPV(pste, cm);
}

//  -------------------------------------------------------------------------

//	This does a null-move search to see if the side not to move has a threat.
//	If they can't drive the score above alpha, even if they get to move twice
//	in a row, their positions probably sucks, and we'll prune this variation.
//	It's not called in endgames, because it has problems detecting zugzwang.

//  We don't need to do a null move if we know allready that there is threats around
//  In endgames any pawn move is a potential threat (score will vary a lot)

static INLINE int FThreat(PCON pcon, PSTE pste, int Alpha, int Beta, const nodeType nt)
{
	bNullMove(pcon, pste);
    // Reduced-depth search
    // R = 3 if (rem_ply > 8) || (rem_ply > 6 (and pieces per side >= 3))
    int R = 3; // set to 3 since we do a verification search

    (pste + 1)->plyRem = pste->plyRem - 1 - R;

    int val;
    if( (pste+1)->plyRem < 0 ) {
        // set plyrem to look for check on next move
        (pste+1)->plyRem = -1;
//        (pste+1)->plyRem = 0;
        val = -SearchQ(pcon, pste + 1, - Beta, - Beta + 1, true);
    } else {
        val = -Search(pcon, pste + 1, - Beta, - Beta + 1, nt);
    }
    bUndoNullMove(pcon, pste);
    return val;
}


//  -------------------------------------------------------------------------

static int INLINE capture_value(PCON pcon, const STE *pste, const CM *pcm) {
    if( moveisPromotion( pcm ) )
        return valKNIGHT - valPAWN;
    int pc = Pc(pcm->m);
    if( (pc|1) == (B_EP|1) )
        return 0;
    int took = Capt(pcm->m);
    if( exchangePieceValue[ took ] >= exchangePieceValue[ pc ] )
        return exchangePieceValue[ took ] - exchangePieceValue[ pc ];

    // compute see value...
    int val = see(pcon, pste, From(pcm->m), To(pcm->m));
    return val;
}

#define NCMove(from, to, pc) ((from) | ((to) << 6) | ((pc) << 12))
#define CMove(from, to, type, capt) ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16))

//	Quiescent search.  This lets captures play out at the end of the search,
//	so I don't end up aborting a line at a stupid place.

//#define QS_MARGIN   75
//#define QS_MARGIN   55
//#define QS_MARGIN   95
#define QS_MARGIN   150

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta, bool do_check)
{
	pcon->ss.nodes++;
    stats.QS_count ++;
	CM cmBest(0,0);
	pste->ccmPv = 0;
	bool fFound = false;
    bool isPV = Beta != (Alpha+1);
    int iaminCheck = pste->checkf;
    int  f_max   = valMAX;
    CM * p_moves;
    int depth = int(pste - pcon->argste) + 1;

    if (FRepSet(pcon, pste, pste->hashkPc)) {
        stats.repCount++;
		return 0;
	}

#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#endif
    int OldAlpha = Alpha;
	//
	//	See if static eval will cause a cutoff or raise alpha.
	//
    // this will be the return value if king evasion does no capture
	int posVal = ValEval(pcon, pste, Alpha, Beta);
    // score is valMIN when in check or else it won't play moves that lead to loss of material when nearly mated
    int bestVal = valMIN;

    if( depth >= (csteMAX-2) )
        return posVal;

    // generate captures, if we are in check then generate all moves (evade)
	if (iaminCheck) {
		bGenEvasions(pcon, pste);
        pste->plyRem++;
        do_check = do_check || QS_CHECK_LIMIT;
	}
	else {
#ifdef PRUNE_QS
        f_max = posVal + QS_MARGIN;
#endif
        bestVal = posVal;
	    if (bestVal >= Beta)
		    return bestVal;
        if (bestVal > Alpha)
            Alpha = bestVal;
		p_moves = bGenCaptures(pcon, pste, pste->mlist_start);
		p_moves = bGenPasserPush(pcon, pste, p_moves);
	}

    if( depth > pcon->ss.plyMaxDepth )
        pcon->ss.plyMaxDepth = depth;
    PrepareSortQS( pcon, pste, depth );
    //
	//	Iterate through capturing moves, to see if I can improve upon alpha
	//	(which may be the static eval).
	//

    foreachmove(pste, pcm) {

        SortNext(pste, pcm);
        fFound = true;
        bMove(pcon, pste, pcm->m);
#ifdef PRUNE_QS
        if( !iaminCheck
//            && (!isPV)
            // end game values
            && ((pste+1)->valPcUs >= valROOK)
            && ((pste+1)->valPcThem >= valROOK)
            && !moveisPromotion(pcm)
            && !moveisPawnMove(pcm)
            ) {
                if( moveisBadCapture(pcm) ) {
    		        bUndoMove(pcon, pste, pcm->m);
//                    break;
                    continue;
                }
                if( pste[1].checkf == 0 || (!do_check) ) {
                    int approx = f_max + MaterialGain(pcon, pcm);
 /*                   if( moveisPromotion(pcm) ) {
                        if( Prom(pcm->m) == PROM_Q )
                            approx += valQUEEN;
                        else if( Prom(pcm->m) == PROM_R )
                            approx += valROOK;
                        else if( Prom(pcm->m) == PROM_B || Prom(pcm->m) == PROM_N )
                            approx += valBISHOP;
                    }*/
                    if ( (!isPV) && approx <= Alpha ) {
                        stats.FP_QS++;
    		            bUndoMove(pcon, pste, pcm->m);
                        if( approx > bestVal ) {
                            cmBest.m = 0;
	                        pste->ccmPv = 0;
                            bestVal = approx;
                        }
                        continue;
                    }
                }
        }
#endif

		(pste + 1)->plyRem = pste->plyRem - 1;
		int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha, QS_CHECK_LIMIT);
		bUndoMove(pcon, pste, pcm->m);

        if (val >= Beta) {
            copyPVQS(pste, *pcm );
            bestVal = val;
            goto out;
        }
        if (val > bestVal) {
			bestVal = val;
            cmBest = *pcm;
            copyPVQS(pste, *pcm );
            if (val > Alpha)
			    Alpha = val;
        }
	}
    int sp = bestVal;   // stand pat score
    if( !iaminCheck && cmBest.m != 0 ) {
        sp = valMIN;
    }
#if 1
//    if( !iaminCheck && do_check && (bestVal < (valKNIGHT/2))
    if( !iaminCheck && do_check && (bestVal < (valKNIGHT))
        && (pste->valPcUs >= valROOK)) {
        // gen checks
        // this is of course not optimal
        //int extCount = depth - pcon->ss.plyDepth + pste->plyRem; // extCount < 0 => reduced or wanted depth not reached yet
        //int maxExt = MAXEXT(depth - extCount);
        //if( !fFound || (extCount < maxExt) ) 
#if 0
/*        int oldcheckCount = 0;
        int newcheckCount = 0;
        char oldcheck[80][10];
        char newcheck[80][10];*/
        {
            p_moves = bGenNonCaptures(pcon, pste, pste->mlist_start);
//            p_moves = bGenPseudoCheck(pcon, pste, pste->mlist_start);

            foreachmove(pste, pcm) {

                fFound = true;
                bMove(pcon, pste, pcm->m);
                // we only want moves that do check
                if( (pste[1].checkf == 0 || moveisPromotion(pcm)) ) {
		            bUndoMove(pcon, pste, pcm->m);
                    continue;
                }
//                CmToSz(pcm, oldcheck[oldcheckCount]);
//                oldcheckCount++;
                // don't do check if it looses material
                if( see(pcon, pste+1, To(pcm->m)) < 0) {
                    bUndoMove(pcon, pste, pcm->m);
                    continue;
                }

		        (pste + 1)->plyRem = pste->plyRem - 1;
		        int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha, QS_CHECK_LIMIT);
		        bUndoMove(pcon, pste, pcm->m);
                if (val >= Beta) {
                    bestVal = val;
                    cmBest = *pcm;
                    copyPVQS(pste, *pcm );
			        goto out;
                }
                if (val > bestVal) {
			        bestVal = val;
                    cmBest = *pcm;
                    copyPVQS(pste, *pcm );
                    if (val > Alpha)
			            Alpha = val;
                }
	        }
        }
#else
        {
//            p_moves = bGenNonCaptures(pcon, pste, pste->mlist_start);
            p_moves = bGenPseudoCheck(pcon, pste, pste->mlist_start);

            foreachmove(pste, pcm) {

                fFound = true;
                bMove(pcon, pste, pcm->m);
                // we only want moves that do check
                if( (pste[1].checkf == 0) ) {
		            bUndoMove(pcon, pste, pcm->m);
                    continue;
                }
                //CmToSz(pcm, newcheck[newcheckCount]);
                //newcheckCount++;
                // don't do check if it looses material
                if( see(pcon, pste+1, To(pcm->m)) < 0) {
                    bUndoMove(pcon, pste, pcm->m);
                    continue;
                }

		        (pste + 1)->plyRem = pste->plyRem - 1;
		        int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha, QS_CHECK_LIMIT);
		        bUndoMove(pcon, pste, pcm->m);
                if (val >= Beta) {
                    bestVal = val;
                    cmBest = *pcm;
                    copyPVQS(pste, *pcm );
			        goto out;
                }
                if (val > bestVal) {
			        bestVal = val;
                    cmBest = *pcm;
                    copyPVQS(pste, *pcm );
                    if (val > Alpha)
			            Alpha = val;
                }
	        }
        }
#endif
//        if( newcheckCount != oldcheckCount )
//            newcheckCount = newcheckCount;

//        if(!fFound)
//            return 0;
    }
#endif
#ifdef QS_VERIFY_EXTENSION
//#define QS_SP_MARGIN    200
#define QS_SP_MARGIN    75
//#define QS_SP_MARGIN    45
    if( fFound && !iaminCheck && (pste->plyRem >= -1) && (sp + 25 > bestVal)
//    if( fFound && !iaminCheck && (pste->plyRem >= -5) && (sp + 25 > bestVal)
        && (bestVal > OldAlpha)
//        && !(pste-1)->fNull   // no recursion
//       && !(pste-2)->fNull     // one recursion allowed
        && (depth <= 4 || !(pste-4)->fNull)     // three recursions allowed
        ) {
        stats.qs_sp++;
        // threat check
	    bNullMove(pcon, pste);
        (pste + 1)->plyRem = pste->plyRem - 1;
        int val = -SearchQ(pcon, pste + 1, - Beta + QS_SP_MARGIN, - Alpha + QS_SP_MARGIN, true);
        bUndoNullMove(pcon, pste);
        if( val + QS_SP_MARGIN <= Alpha ) {
//        if( val + 25 < Alpha ) {
            pste->plyRem = 0;
            pste->fNull = true; // set recursion flag
            pste->reduced = false;
            // extend one ply, note that we are in the PV
            bestVal = SearchPV(pcon, pste, OldAlpha, Beta);
//            bestVal = SearchPV(pcon, pste, val + QS_SP_MARGIN, Beta);
            goto out;
        }
    }
#endif

	if (!fFound) {
        // We have no move !!
		if (iaminCheck) {
            // Checkmate
			bestVal = mateValue(depth);
		}
	}
out:
    Assert(ValIsOk(bestVal));
	return bestVal;
}


//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

#define LOWERBOUND  OldAlpha


static INLINE bool isNodeUseless(CON *pcon, STE *pste, U32 hashMoveNodeCount, PCM pcm) {
//    if( moveisGood(pcm) )
//        return false;
    if( moveisCapture(pcm) )
        return false;

/*    if( (pste->threatMove != U32(-1)) && (To(pste->threatMove) == From(pcm->m)) ) {
        return false;
    }
*/
    return true;

}
#if 0
static void CountStuff(CON *pcon, STE *pste, nodeType nt) {

    if( nt == nodeCUT )
        return;
    if( pste->plyRem <= 1 )
        return;

    return;
    foreachmove(pste,pcm) {
        // sort quiescent moves and 'bad' captures
        U32 mask = pcm->cmk & cmkMASK;
        if( mask <= cmkGOODCAPT ) {

	        bMove(pcon, pste, pcm->m);
            PHASH new_hash_entry = ProbeHashValue(pste+1);
            if( mask == cmkKILLER || mask == cmkQUEEN )
                mask = cmkNONE;
            if( new_hash_entry ) {
                U32 effort = new_hash_entry->count;
                if( effort >= cmkBADCAPT )
                    effort = cmkBADCAPT - 1;
//                    if( - new_hash_entry->val < -valMATEscore )
//                        effort = 0; // look that last

/*                    if( nt == nodeCUT ) {
                    effort = cmkBADCAPT - effort - 1;
                }*/
                pcm->cmk = mask + effort;
            }
            bUndoMove(pcon, pste, pcm->m);
        }
    }
}
#endif

static INLINE void extension_check(const CON *pcon, PSTE pste, const CM *pcm, int &extension, int a, int depth) {

    // The first check is not extended
//    (pste+1)->checkdone[pste->side] = true;
#ifdef EXTEND_CHECK
    // extend everywhere in the tree
    // don't extend first check for stm
//    if( pste->checkdone[pste->side]) 
    {
        // stop extending if I lead the score
#ifdef LESS_CHECK
//    if( (a < valKNIGHT) || isFrontier(pste)) 
//    if( (a < valKNIGHT) && ( (depth) >= pcon->ss.plyDepth / 2) ) 
//    if( (pste->val < valKNIGHT) && ( (depth) >= pcon->ss.plyDepth / 2) ) 
//    if(  ( (depth) >= pcon->ss.plyDepth / 2) ) 
//    if( (a < valKNIGHT) ) //&& (depth > -1) ) 
//    if( (a < valKNIGHT) && ( (depth) >= (pcon->ss.plyDepth / 3) ) 
    if( (a < valKNIGHT) && ( true || (depth) >= (pcon->ss.plyDepth / 4) ) 
        || (pcon->ss.plyDepth <= 5) ) 
#endif
        {
            extension = 1;
            stats.c_extcheck++;
        }
    }
#else
    // only extend on check at the frontier
    if( isFrontier(pste) ) {
        // stop extending if I lead the score
        if( a < valKNIGHT ) {
            extension = 1;
            stats.c_extcheck++;
        }
    }
#endif
}

static INLINE void extension_pawn(const CON *pcon, PSTE pste, const HASHPawn *phashp, const CM *pcm, int &extension, bool &can_reduce, int a, bool isPV) {
    if( moveisPawnMove(pcm) ) {
        // Pawn push to 7th rank
        if( moveto7thRank(pcm) ) {
            pste->danger = true;
            can_reduce = false;
            if( isPV )
                extension = 1;
        // no reduction of pawn moves in end game
        } else if (((pste->valPcUs + pste->valPcThem) <= 1*valQUEEN)) {
            pste->danger = true;
            can_reduce = false;
        } else {
            // don't reduce push of passed pawns
            int from = From(pcm->m);
            bitboard from_64 = IdxToU64(from);
            if( from_64 & phashp->side[pste->side].passed )
                can_reduce = false;
        }
    } else if( moveisPromotion(pcm) )
        can_reduce = false;
}

static INLINE void extension_castle(const CON *pcon, PSTE pste, const CM *pcm, int &extension, bool &can_reduce, int a, bool isPV, bitboard kingRing) {
#if 0
    if( ((Capt(pcm->m)|1) == (PAWN|1)) && (kingRing & IdxToU64( To(pcm->m) )) ) {
        pste->danger = true;
        can_reduce = false;
//        if( isPV )
            extension = 1;
    }
#else
    #ifndef MINIMIZE
        if( (kingRing & IdxToU64( To(pcm->m) )) ) {
            pste->danger = true;
            can_reduce = false;
        }
    #endif
#endif
}

static INLINE void before_search(const CON *pcon, PSTE pste, bool isPV, bitboard &kingRing, bool & forced, int & preExtend, int extCount, int depth) {

    U8 otherside = FlipColor(pste->side);
    kingRing = king_attacks[pcon->king[otherside]];
    if( otherside == B )
        kingRing |= soutOne(kingRing);
    else
        kingRing |= nortOne(kingRing);
    if( -5*(extCount-1)/2 >= pcon->ss.plyDepth ) {
//        forced = true;
    }
#ifndef MINIMIZE
    if( !pste->checkf && pste->evaluated && !pste->evalislazy && (pste-1)->evaluated && !(pste-1)->evalislazy
        && (pste->plyRem >= 1)
        && (pste)->prevCaptValue == 0) {
            if( abs(pste->val +(pste-1)->val) >= 50 ) {
                forced = true;
            }
        }
#endif
#if 1
    Killers[depth+2+2].raz();
    MateKillers[depth+2+2].raz();
//    Killers[depth+2].raz();
//    MateKillers[depth+2].raz();
#endif

}


static inline hashBound calcBoundType(int val, int Alpha, int Beta) {
    if( val >= Beta )
        return hashfBETA;
    else if( val <= Alpha )
        return hashfALPHA;
    else
        return hashfEXACT;
}

static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, nodeType nt)
{
	int	val;
    pste->ccmPv = 0;
    pste->reduced = false;
	CM cmBest(0,0);
    bool king_attak = false;
    int preExtend = 0;
    int bestVal = valMIN;
    int iaminCheck = pste->checkf;
    bool nm_threat = false;
    // the condition is no more true !! because of lmr re-search on pv sibblings
    Assert(Beta == (Alpha+1));

    if (pste->plyRem < 0) {
        return SearchQ(pcon, pste,  Alpha, Beta, true);
    }

    int depth = int(pste - pcon->argste) + 1;
    if( depth >= (csteMAX-5) ) {
        return SearchQ(pcon, pste,  Alpha, Beta, true);
    }

    //	Increment global node counter, and figure out if it is time to call
    //	the interface callback, which is what allows for an interruptible
    //	search.  It's crucial that this be called now and then.
    //
    if (++pcon->ss.nodes >= pcon->ss.nodesNext) {
        //
        //	20000 nodes is maybe 1/10 to 1/30 of a second.
        //
        pcon->ss.nodesNext = pcon->ss.nodes + 20000;
        VPrCallback(pcon);
        VCheckTime(pcon);
        if ((pcon->fAbort) || (pcon->fTimeout))
            return 0;
        if( pcon->smode == smodeTHINK || pcon->smode == smodeANALYZE )
            VPrSendUciCurrentLine(pcon, pste);
	}
    //
    //	I can return draws and cut off if I'm somewhere other than the
    //	first ply of depth.
    //
    //	Return draw score if this is a repeated node.
    //
    if (depth > 2 && FRepSet(pcon, pste, pste->hashkPc)) {
        stats.repCount++;
        return 0;
    }

#ifdef EVAL_INTERIOR
    // gathering stats
    int nmPosVal = pste->val;
    if(!pste->evaluated )
	    nmPosVal = ValEval(pcon, pste, Alpha, Beta);
#else
    int nmPosVal = MaterialValue(pste) + valPAWN;
#endif

    // extCount is the net number of extension done (all extensions - all reductions done)
    int extCount = depth - pcon->ss.plyDepth + pste->plyRem; // extCount < 0 => reduced or wanted depth not reached yet
    int maxExt = MAXEXT(depth - extCount);

    //	Check hash table, in order to get "best" moves, and try to cut
    //	off.  Don't cut off within two plies of the root, otherwise the
    //	program might allow stupid rep draws.
    //
    PHASH hash_entry = 0;
    bool forced = false;
//    U32 hashMoveNodeCount = 0;

#ifdef USE_HASH
    hash_entry = ProbeHashValue(pste);
//    int hash_score = (hash_entry ? hash_entry->val : valMAX);
//    if( hash_entry )
//        hashMoveNodeCount = hash_entry->count;

    if( hash_entry && (hash_entry->plyRem >= pste->plyRem) && (depth > 2) ) {
        PHASH phash = hash_entry;
        int val = phash->val;
#if 0
        if( val >= valMATEscore ) {
            val -= depth - 1;
        } else if( val <= -valMATEscore ) {
            val += depth;
        }
#endif
		switch(phash->getBound())
        {
			case hashfEXACT:
                stats.Hcutoff ++;
                return val;
                break;
			case hashfBETA:
                if( val >= Beta ) {
                    stats.Hcutoff ++;
					return val;
                }
				break;
			case hashfALPHA:
                if( val <= Alpha ) {
                    stats.Hcutoff ++;
					return val;
                }
				break;
            default:
                hash_entry = 0;
                //Assert(0);
                break;
        }
	}
#endif
    if( hash_entry )
        nm_threat = hash_entry->getThreat();

	//
	// Check alpha/beta window against a mate score
    // Thx for the idea
	//
    int OldAlpha = Alpha;
#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#if 1
    // this is the best score we can get from here
    int i_will_mate_score =  -mateValue(depth);
    if( i_will_mate_score < Beta ) {
        Beta = i_will_mate_score;
        // if I set + 0 then I don't see the final move (
            // I will not see something better than Alpha so just return
        // Alpha is mate in n, best here is mate in n+m, so return
        if( i_will_mate_score <= Alpha )
            return Alpha;
//            return i_will_mate_score;
    }
#endif
#endif

    //
    // try recognizers
    //
#ifdef USE_RECOGNIZER
    // the recognizers of bitbases where allready probed in the eval
    if( true
        && (pste->bound != fail_bound) && pste->evaluated) {
        recog_result ret = {recog_bound(pste->bound), pste->val, recog_allways_valid};
        if( ret.bound == exact_bound )
            return ret.score;
        if( ret.bound == lower_bound ) {
            if( ret.score >= Beta )
                return ret.score;
//            if( ret.score > Alpha )
//                Alpha = ret.score;
        }
        if( ret.bound == upper_bound ) {
            if( ret.score <= Alpha )
                return ret.score;
            if( ret.score < Beta )
                Beta = ret.score;
        }
    }
#endif

    PHASHPawn phashp;
    EvalPawns(pcon, pste, phashp);

    static int deltaRazor[] = {150,225,300};

    pste->threatMove = U32(-1);
    pste->threatTarget = 0;
#ifdef NULL_MOVE
	//
	//	Do null-move search, which is a reduced depth that tries to
	//	figure out if the opponent has no threats.
	//
    // Note :
    // - if Beta is a mate score then it is not probable that I can do better without doing anything
    // - only done if not a PV (what about branches that will become a pv later ?)
    // - nm_threat caching does not work for an unknown reason (more nodes sometimes)
    // - material constraint reduced, I hope this works well with the verification search
    // - reductions are now allowed after an extension (so no test on extCount)
    // - if the stand pat score is < Beta then doing nothing will not raise it to Beta or more
    if (
        (!iaminCheck)
        && (pste->valPcUs > 0)
//        && (pste->valPcThem > 0)
        && (nmPosVal >= Beta)
        && (Beta < valMATEscore)    // redondant with above test
        && (pste->plyRem >= 1)      // need more tests
//        && (pste->plyRem >= 2)      // need more tests
        && (pcon->ss.plyDepth > 4)  // no null move in the firsts search
#ifdef TWO_NULL
        && (!( pste->fNull && (pste-1)->fNull) )
#else
        && (!pste->fNull)
#endif
        ) {
        {
            stats.NM1++;
            // don't forget to eval when the call to eval will be after bMove()
            int val = FThreat(pcon, pste, Alpha, Beta, nt);
            Assert(ValIsOk(val));
            if( val >= Beta ) {
                // why not store result in hash ?
                // TODO check this: this does not work well on some position (more nodes)
                stats.NM2++;
                // verification search (for end game mostly)
                // still need to do NM at lower depth (higher remaining depth)
                const int Rv = 1 + 4;
                if(true && pste->plyRem >= Rv) {
                    stats.NM3++;
                    pste->fNull = true; // set or else we will loop :)
                    pste->plyRem -= Rv;
                    val = Search(pcon, pste, Alpha, Beta, nodeCUT /*nt*/);
                    pste->plyRem += Rv;
                }
            }
            if( val >= Beta ) {
#ifdef HASH_NM
                // store because we don't want to do this same search *again*
                RecordHashValue(pcon, pste, cmBest, val/*Beta*/, Alpha, Beta, depth, false);
#endif
                return val;
            }
            pste->threatMove = (pste+1)->argcmPv[0].m;
            if(pste->threatMove)
                pste->threatTarget = pcon->pos[ To(pste->threatMove) ];
#ifdef BM_EXTENSION
            if( true && (pste->plyRem <= BM_DEPTH) && pste->threatMove 
                && (val + valPAWN + 50 < Alpha 
//                && Alpha > -valKNIGHT
                )
                && (((pste->threatTarget == (pste-2)->threatTarget))
                    || To(pste->threatMove) == To((pste-2)->threatMove))
                && !( (To((pste-2)->lastMove) == To(pste->threatMove))
                    || (To((pste-1)->lastMove) == From(pste->threatMove)))
                 ){
//                    preExtend = 1;
//                    king_attak = true;
                    stats.c_bmext++;
/*                    pste->lastMove = pste->threatMove; // for debug only
                    char tmp[256];
                    sprintf(tmp, "same threat...  rem=%d targ=%d %x %x %d [%d]", pste->plyRem, pste->threatTarget, To(pste->threatMove), To((pste-2)->threatMove), val, Alpha);
                    VDumpLine(pcon, pste, tmp);
                    if( (pste-1)->reduced )
                        printf(" and father is reduced\n");
                    if( (pste-2)->reduced )
                        printf(" and father 2 is reduced\n");
                    if( (pste-3)->reduced )
                        printf(" and father 3 is reduced\n");*/
                    if( (pste-1)->reduced )
                        return valMIN;
            } else
#endif
            if( false && val < nmPosVal - 75 ) {
                if( pste->threatTarget == 0 ) {
                    bool threat = true;
                    preExtend = 1;
                }
                else if ( val + 75 < nmPosVal - exchangePieceValue[ pste->threatTarget ] ) {
                    bool threat = true;
                    preExtend = 1;
                }
            }
            else if( true && (val < Alpha)
//                && (pste->plyRem <= 7)
                // there will be a re-search
                && (pste-1)->reduced
                && !( moveisVeryGoodCapture(&(pste+1)->argcmPv[0])
                    || (To((pste-2)->lastMove) == To(pste->threatMove))
                    || (To((pste-1)->lastMove) == From(pste->threatMove))
                    ) ) {
                    return valMIN;
            }
            else if( false && (val < Alpha)
                && (pste->plyRem <= 5)
                // there will be a re-search
                && (pste-1)->reduced
//                && !moveisGoodCapture(&(pste+1)->argcmPv[0])
                && !moveisEqualCapture(&(pste+1)->argcmPv[0])
                && ( 
                    (To((pste-1)->lastMove) == From(pste->threatMove))
                    || (From((pste-1)->lastMove) == To(pste->threatMove))
                     )) {
//                    pste->lastMove = pste->threatMove; // for debug only
//                     VDumpLine(pcon, pste, "conn:");
                    return valMIN;
            }
            else if( true && val <= -valMATEscore ) {
                // extend one ply if there is a mate threat
                // ok storing threat in hash does not work for some reason...
                // only setting king_attack so that we don't reduce the search
                // if we extend then we can search a lot deeper for no reason when
                // there is a mate but winning side does not play the winnin move
                if( (pste-1)->reduced )
                    return valMIN;
                //king_attak = true;
                // threat flag should be stored in hash too for next search
            }
        }
	}
#ifdef RAZORING
    // Razoring: la Glaurung
//    else if( true && (pste->plyRem <= 2) && (nmPosVal + 250 + pste->plyRem*150 <= Alpha) && (Alpha < valMATEscore)
//    else if( true && ((pste->plyRem <= 1) || (pste->plyRem <= 2 && extCount*3 >= depth)) && (nmPosVal + 350 <= Alpha) && (Alpha < valMATEscore)
//    else if( true && ((pste->plyRem <= 2) ) && (nmPosVal + 350 <= Alpha) && (Alpha < valMATEscore)
//    else if( true && (pste->plyRem <= 2) && (nmPosVal + 175 + pste->plyRem*175 <= Alpha) && (Alpha < valMATEscore)
    else if( true && (pste->plyRem <= 2) && (nmPosVal + deltaRazor[pste->plyRem] <= Alpha) && (Alpha < valMATEscore)
//    else if( true && (pste->plyRem <= 2) && (nmPosVal + 350 <= Alpha) && (Alpha < valMATEscore)

        && (nt == nodeALL)
//        && (pste->valPcThem > 0)

        ) {
        stats.c_razor++;
        int d = pste->plyRem;
//        pste->plyRem = -1;
//            pste->plyRem = d+1;
        int val = SearchQ(pcon, pste, Alpha, Beta, true);
        if(val <= Alpha)
            return val;
        stats.c_razor_nok++;
        pste->plyRem = d;
    }
#endif
#endif


    //	Generate moves.
    //
    bGenMoves(pcon, pste);
    //
    //	I know how many reversible moves have been made up until now.
    //	I just tried to make another move.  If I am here, I didn't
    //	leave myself in check, so I succeeded.
    //
    //	If the number of reversible moves made before this is at
    //	least 50 for each side, this was a draw.
    //
    int lmc = legalMoveCount(pste);

    if ( (pste->plyFifty >= 50 * 2) && (lmc > 0) ) {
        pste->ccmPv = 0;
        return 0;
    }
    // single move extension
    if( (lmc == 1) && iaminCheck)
        preExtend = 1;

    bitboard kingRing = 0;
    before_search(pcon, pste, /*isPV*/ false, kingRing, forced, preExtend, extCount, depth);

    PrepareSort( pcon, pste, depth, phashp );

	//	Mark the hash moves so it sorts first.
	//
#ifdef USE_HASH
	if (hash_entry != NULL)
		SortHash(pste, hash_entry);
#endif
#ifdef USE_KILLERS
    SortKillers(pste, depth);
#endif

    int a = Alpha;
    int b = Beta;

    unsigned int prevCaptureSqr = pste->prevCaptSqr;

    bool fFound = false;
    int  c_move = 0;

    int  f_margin = valMAX;
    bool can_prune = false;

#define FUTILITY_MARGIN (150)

#ifdef FUTILITY_PRUNING
    if( !iaminCheck
//        && ((pste->valPcUs + pste->valPcThem) >= 2*valQUEEN)
        && ((pste->valPcUs + pste->valPcThem) >= 1*valQUEEN)
//        && !phashp->pawnon6or7
        && (pste->plyRem <= 1) )
        can_prune = true;

    f_margin = FUTILITY_MARGIN * (pste->plyRem + 1);
#endif

//    pste->ttmove = hash_entry && (hash_entry->getMove() != 0);
//    int altMoves = 0;

//    CountStuff(pcon, pste, nt);

    //
    //	Iterate through the move list, trying everything out.
    //
    foreachmove(pste,pcm) {
        //
        //	All moves are sorted on the cmk field
        //	cmk is set in PrepareSort() and in other sort functions

        SortNext(pste, pcm);
        //
        //	Make the move

        bMove(pcon, pste, pcm->m);
        stats.c_int++;
	    //int posVal = ValEval(pcon, pste, a, b);

        int extension = preExtend;
        // Note : now I allow the reduction after the extension,
        // this is needed for the check extension
        bool can_reduce = (extension) > 0 ? false : true;
        can_reduce = can_reduce && !king_attak;
        can_reduce = can_reduce && !forced;

        if( (pste + 1)->checkf ) {
            can_reduce = false;
            extension_check(pcon, pste, pcm, extension, a, depth);
        }
        extension_pawn(pcon, pste, phashp, pcm, extension, can_reduce, a, /*isPV*/false);

#ifdef RECAPTURE_EXTENSION
        if( Capt(pcm->m) && (To(pcm->m) == prevCaptureSqr)
            && moveisVeryGoodCapture(pcm)
            && (a < valKNIGHT )
            && false
            ) {
                //pste->danger = true;
                can_reduce = false;
                stats.recapture++;
                extension = 1;
        }
#endif
//        if( Capt(pcm->m) ) {
//            extension_castle(pcon, pste, pcm, extension, can_reduce, a, /*isPV*/ false, kingRing);
//        }
        //  ==========================================================================
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( extension && (extCount < maxExt) ) {
            (pste + 1)->plyRem++;
        }
        //
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType = nodeCUT;
        if(nt == nodeCUT )
            succNType = nodeALL;

#ifdef FUTILITY_PRUNING

        if( fFound
            && can_reduce && can_prune
            && !moveisPromotion(pcm)
            && !moveisCapture(pcm)
            ) {
//            int approx = posVal + f_margin;
            int approx = nmPosVal + f_margin;
            if( approx <= a ) {
                stats.FP_frontier++;
                if( approx > bestVal) {
                    bestVal = approx;
                    cmBest.m = 0;
                    pste->ccmPv = 0;
                }
                bUndoMove(pcon, pste, pcm->m);
                c_move++;
                continue;
            }
        }
#endif

        U64 thisMoveNodeCount = pcon->ss.nodes;

#ifdef USE_LMR
        // Note : LMR done on all non pv nodes moves
        //
        // move order is [H] [good capt] [equ capt] [killers] others [bad capt]
        if(fFound
//            && (c_move >= 1)
#if 1
            && (nt == nodeALL)
            && (c_move >= 4)
#else
            && (c_move >= 4)
#endif
//            && (depth >= 3)
//            && (depth >= 2)
            && can_reduce
            && !iaminCheck
//            && (pste->plyRem >= 1)
            && (pste->plyRem >= 2)
            && (a > -valMATEscore)
            && !moveisPromotion(pcm)
            && isNodeUseless(pcon, pste, 0/*hashMoveNodeCount*/, pcm)
            ) {
            stats.LMRhint++;
            // search with a reduced depth
            stats.LMR++;
            pste->reduced = true;

		    (pste + 1)->plyRem --;
        }
#endif
//        val = -Search(pcon, pste + 1, - b, - a, succNType);

        val = -Search(pcon, pste + 1, - Beta, - a, succNType);

        if( pste->reduced ) {
            pste->reduced = false;
            if( val > a ) {
        		(pste + 1)->plyRem ++;
                // re-search at normal depth
                stats.LMRreSearch++;
                val = -Search(pcon, pste + 1, - Beta, - a, succNType);
            }
        }

/*    if( pste->plyRem >= 2 ) {
        thisMoveNodeCount = pcon->ss.nodes - thisMoveNodeCount;
        PHASH hash_entry = ProbeHashValue(pste+1);
        if( hash_entry ) {
            hash_entry->count += U32(thisMoveNodeCount);
        }
    }*/

/*    thisMoveNodeCount = pcon->ss.nodes - thisMoveNodeCount;
    if( (pcon->ss.plyDepth >= 11) && (thisMoveNodeCount > 100000) ) {
        char tmp[256];
        sprintf(tmp, "#> fat node : d=%d n=%I64u ", pcon->ss.plyDepth+1, thisMoveNodeCount);
        VDumpLine(pcon, pste, tmp);
    }
*/
/*    if( (depth > 25)  ) {
        char tmp[256];
        sprintf(tmp, "# long pv : d=%d [%d %d] %d", depth, a, Beta, val);
        VDumpLine(pcon, pste, tmp);
    }*/

        bUndoMove(pcon, pste, pcm->m);
        if ((pcon->fAbort) || (pcon->fTimeout)) {
            return 0;
        }
        //	We got a value back.  We unmade the move.  We're not dead.  Let's
        //	see how good this move was.  If it was >= "beta", it was so good
        //	that we don't need to search for anything better, so we'll leave.
        //
        Assert(ValIsOk(val));
        if( val > bestVal ) {
            bestVal = val;
            if (val > a) {
                cmBest = *pcm;
                copyPV( pste, cmBest );
                if (val >= Beta) {
                    //
                    //	This move failed high
                    //
                    //  Set the new killer move only if it was not marked as good allready
                    //  'Good' moves are allready sorted first so storing them in the killer
                    //  array is counter productive
                    // This was added to see mate search ; with the mate distance bound
                    // no pv was recorded
                    goto cutNode;
                }
                //	This move is between alpha and beta, which is actually pretty
                //	rare.  If this happens I have to add a PV move, and append the
                //	returned PV to it
                //
//                if( nt == nodeALL)
//                    nt = nodeCUT;
                a = bestVal;
            }
        }
        fFound = true;
        c_move++;
//        pste->ttmove = false;

#ifdef NULL_WINDOW
        b = a + 1;              // set the new null window
#endif
//        if( nt == nodeCUT && (bestVal < Alpha))
        nt = nodeALL;
    }

    if (!fFound) {
        // We have no move !!
        if (iaminCheck) {
            // Checkmate
            bestVal = mateValue(depth);
        } else {
            // Draw
            bestVal = 0;
        }
        // ==================================
        return bestVal;
        // ==================================
    }
    // if I have a move, then this is not an all nodes
cutNode:;

    if ((pcon->fAbort) || (pcon->fTimeout)) {
        return 0;
    }

    if( bestVal > Alpha ) {
        pste->lastMove = cmBest.m;
        // for nm refutation
        (pste+1)->lastMove = cmBest.cmk;

        if( cmBest.cmk & cmkHASH )
            stats.c_cut_H++;
        else if( cmBest.cmk & cmkKILLER )
            stats.c_cut_killer++;
        else if( moveisGoodCapture(&cmBest) )
            stats.c_cut_capt++;
        else if( cmBest.cmk & cmkBADCAPT )
            stats.c_cut_badcapt++;
        else
            stats.c_cut_quies++;

        if( bestVal >= Beta ) {
            stats.cCUT++;
            if(iMove == 0)
                stats.cCUTrank1++;
            if(iMove < 5)
                stats.cCUTrankn++;
        } else
            stats.c_alpha_cut++;

        History.update( &cmBest, depth );

#ifdef USE_KILLERS
        // TODO:move mate killer code above, it is never executed here !!
        if( bestVal >= valMATEscore )
            MateKillers[depth].set(&cmBest);
        else if( !moveisCapture(&cmBest) )
            Killers[depth].set(&cmBest);
#endif

//        hashBound hb = calcBoundType(bestVal, OldAlpha, Beta);
//        if( hb != hashfBETA )
//            hb = hb;
//        RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, nm_threat);
        RecordHashValue(pcon, pste, cmBest, bestVal, hashfBETA, depth);
    } else {
        stats.c_all++;
//        hashBound hb = calcBoundType(bestVal, OldAlpha, Beta);
//        if( hb != hashfALPHA )
//            hb = hb;
        RecordHashValue(pcon, pste, cmBest, bestVal, hashfALPHA, depth);
//        RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, nm_threat);
    }


    Assert(ValIsOk(bestVal));
    return bestVal;
}



static int SearchPV(PCON pcon, PSTE pste, int Alpha, int Beta)
{
	int	val;
    pste->ccmPv = 0;
    pste->reduced = false;
	CM cmBest(0,0);
    int preExtend = 0;
    int bestVal = valMIN;
    int iaminCheck = pste->checkf;

    if (pste->plyRem < 0) {
        return SearchQ(pcon, pste,  Alpha, Beta, true);
    }

    int depth = int(pste - pcon->argste) + 1;
    if( depth >= (csteMAX-5) ) {
        return SearchQ(pcon, pste,  Alpha, Beta, true);
    }

	//	Increment global node counter, and figure out if it is time to call
	//	the interface callback, which is what allows for an interruptible
	//	search.  It's crucial that this be called now and then.
	//
	if (++pcon->ss.nodes >= pcon->ss.nodesNext) {
		//
		//	2000 nodes is maybe 1/50to 1/100 of a second.
		//
		pcon->ss.nodesNext = pcon->ss.nodes + 10000;
		VPrCallback(pcon);
		VCheckTime(pcon);
        if( pcon->smode == smodeTHINK || pcon->smode == smodeANALYZE )
            VPrSendUciCurrentLine(pcon, pste);
	}
	if ((pcon->fAbort) || (pcon->fTimeout))
		return 0;
	//
	//	I can return draws and cut off if I'm somewhere other than the
	//	first ply of depth.
	//
	//	Return draw score if this is a repeated node.
	//
	if (depth > 2 && FRepSet(pcon, pste, pste->hashkPc)) {
        stats.repCount++;
		return 0;
	}

    // gathering stats
    int nmPosVal = pste->val;
    if(!pste->evaluated )
	    nmPosVal = ValEval(pcon, pste, Alpha, Beta);

    // extCount is the net number of extension done (all extensions - all reductions done)
    int extCount = depth - pcon->ss.plyDepth + pste->plyRem; // extCount < 0 => reduced or wanted depth not reached yet
    int maxExt = MAXEXT(depth - extCount);

	//	Check hash table, in order to get "best" moves, and try to cut
	//	off.  Don't cut off within two plies of the root, otherwise the
	//	program might allow stupid rep draws.
	//
    PHASH hash_entry = 0;
#ifdef USE_HASH
//    U32 hashMoveNodeCount = 0;
    hash_entry = ProbeHashValue(pste);
//    if( hash_entry )
//        hashMoveNodeCount = hash_entry->count;

#endif

	//
	// Check alpha/beta window against a mate score
    // Thx for the idea
	//
    int OldAlpha = Alpha;
#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#if 1
    // this is the best score we can get from here
    int i_will_mate_score =  -mateValue(depth);
    if( i_will_mate_score < Beta ) {
        Beta = i_will_mate_score;
        // if I set + 0 then I don't see the final move (
            // I will not see something better than Alpha so just return
        // Alpha is mate in n, best here is mate in n+m, so return
        if( i_will_mate_score <= Alpha )
            return Alpha;
//            return i_will_mate_score;
    }
#endif
#endif

    //
    // try recognizers
    //
#ifdef USE_RECOGNIZER
    // the recognizers of bitbases where allready probed in the eval
    if( true && (depth > 4)
        && (pste->bound != fail_bound) && pste->evaluated) {
        recog_result ret = {recog_bound(pste->bound), pste->val, recog_allways_valid};
        if( ret.bound == exact_bound )
            return ret.score;
        if( ret.bound == lower_bound ) {
            if( ret.score >= Beta )
                return ret.score;
        }
        if( ret.bound == upper_bound ) {
            if( ret.score <= Alpha )
                return ret.score;
            if( ret.score < Beta )
                Beta = ret.score;
        }
    }
#endif
    pste->threatMove = U32(-1);
    pste->threatTarget = 0;
#ifdef BM_EXTENSION
    if (
        (!iaminCheck)
        && (pste->valPcUs > 0)
        && (pste->valPcThem > 0)
        && (pste->plyRem <= BM_DEPTH)
        && (pste->plyRem >= 1)
        && (!pste->fNull)) {
        if( true ) {
            stats.NM1++;
            int bound = nmPosVal - 75;
            int val = FThreat(pcon, pste, bound-1, bound, nodePV);
            Assert(ValIsOk(val));
            pste->threatMove = (pste+1)->argcmPv[0].m;
            if(pste->threatMove)
                pste->threatTarget = pcon->pos[ To(pste->threatMove) ];
            if( true && (val + valPAWN < nmPosVal && pste->threatMove)
                && (pste->threatTarget == (pste-2)->threatTarget
                    || To(pste->threatMove) == To((pste-2)->threatMove))
                && !( (To((pste-2)->lastMove) == To(pste->threatMove))
                    || (To((pste-1)->lastMove) == From(pste->threatMove)))
                 ){
//                    preExtend = 1;
//                    king_attak = true;
                    stats.c_bmext++;
//                    pste->lastMove = pste->threatMove; // for debug only
//                    VDumpLine(pcon, pste, "PV:some threat... ");
            }
        }
	}
#endif

    PHASHPawn phashp;
    EvalPawns(pcon, pste, phashp);

#ifdef USE_HASH
#ifdef USE_IID
    if( (hash_entry == NULL || (hash_entry->getMove() == 0)) && (pste->plyRem >= (4)) ) {
//    if( (hash_entry == NULL || (hash_entry->getMove() == 0)) && (pste->plyRem >= (3)) ) {
        // depth /2 gives worse results (more nodes)
//        int reduc = pste->plyRem / 2;
        int reduc = 2;
        pste->plyRem -= reduc;
        stats.c_iid++;
        int iidval = SearchPV(pcon, pste, Alpha, Beta);
#if 1
        /* frombeowulf : Re-search properly if the previous search failed low, so that we know we're getting
         * a good move, not just the move with the highest upper bound (which is essentially
         * random and depends on the search order.) */
        if( (iidval <= Alpha) ) {
//            iidval = SearchPV(pcon, pste, iidval - valPAWN, Beta);
            iidval = SearchPV(pcon, pste, valMIN, Beta);
            stats.c_iidbad++;
        }
#endif
        pste->plyRem += reduc;
        hash_entry = ProbeHashValue(pste);
        // Is it possible to not have a hash move here ? (how many times ?)
    }
#endif
#endif
    bool forced = false;

	//	Generate moves.
	//
    bGenMoves(pcon, pste);
	//
	//	I know how many reversible moves have been made up until now.
	//	I just tried to make another move.  If I am here, I didn't
	//	leave myself in check, so I succeeded.
	//
	//	If the number of reversible moves made before this is at
	//	least 50 for each side, this was a draw.
	//
    int lmc = legalMoveCount(pste);
    int rankCut = lmc;
	if ( (pste->plyFifty >= 50 * 2) && (lmc > 0) ) {
		pste->ccmPv = 0;
		return 0;
	}
    // single move extension
    if( (lmc == 1) && iaminCheck)
        preExtend = 1;

    bitboard kingRing = 0;
    before_search(pcon, pste, /*isPV*/ true, kingRing, forced, preExtend, extCount, depth);

    PrepareSort( pcon, pste, depth, phashp );
	//	Mark the PV and hash moves so they sort first.
	//
    // this should not be needed anymore
	if (lastPV[depth].m != 0) {
		SortPV(pste, depth);
		lastPV[depth].m = 0;
	}
#ifdef USE_HASH
	if (hash_entry != NULL)
		SortHash(pste, hash_entry);
#endif
#ifdef USE_KILLERS
    SortKillers(pste, depth);
#endif

    int a = Alpha;
    int b = Beta;

    unsigned int prevCaptureSqr = pste->prevCaptSqr;

//    int altMoves = 0;
    bool fFound = false;
    int  c_move = 0;
//    pste->ttmove = hash_entry && (hash_entry->getMove() != 0);
#if 0
    {
        // we will sort the moves on their real score : get it from the hash table or do some search
        // this is an iid on all moves
        foreachmove(pste,pcm) {
            // sort quiescent moves and 'bad' captures
            U32 mask = pcm->cmk & cmkMASK;
//            if( mask <= cmkKILLER ) {
            if( mask <= cmkGOODCAPT ) {

		        bMove(pcon, pste, pcm->m);
                PHASH new_hash_entry = ProbeHashValue(pste+1);
                if( mask == cmkKILLER || mask == cmkQUEEN )
                    mask = cmkNONE;
                if( new_hash_entry ) {
                    int val = - new_hash_entry->val;
                    if( (pste+1)->checkf )
                        val -= 5;
                    pcm->cmk = mask + valMAX + val;
                } else if( true && pste->plyRem >= 4 ) {    // 2
                    // >= 4 - 2
                    // >= 2 / 2
                    // >= 3 - 3
//                    (pste+1)->plyRem = pste->plyRem / 2;
#if 1
                    (pste+1)->plyRem = pste->plyRem - 3;
                    // full windows does not give a lot more nodes, TT filled with 'good' values
                    int val = -SearchPV(pcon, pste + 1, -b, -a);
#else
//                    (pste+1)->plyRem = 0;
//                    int val = -SearchPV(pcon, pste + 1, -b, -a);
                    int val = -SearchQ(pcon, pste + 1, -b, -a, true);
#endif
//                    val = -Search(pcon, pste + 1, beta_lim - 100, beta_lim, nodeCUT);
                    if( (pste+1)->checkf )
                        val -= 5;
                    pcm->cmk = mask + valMAX + val;
                }
                bUndoMove(pcon, pste, pcm->m);
            }
        }
    }
#else
//    if( pste->plyRem >= 1 )
    if( false )
    {
        // we will sort the moves on their real score : get it from the hash table or do some search
        // this is an iid on all moves
        foreachmove(pste,pcm) {
            // sort quiescent moves and 'bad' captures
            U32 mask = pcm->cmk & cmkMASK;
            if( mask < cmkGOODCAPT ) {
//            if( mask <= cmkEQUCAPT ) {
		        bMove(pcon, pste, pcm->m);
                PHASH new_hash_entry = ProbeHashValue(pste+1);
                if( mask == cmkKILLER || mask == cmkQUEEN )
                    mask = cmkNONE;
                if( new_hash_entry ) {
                    U32 effort = new_hash_entry->count;
                    if( effort >= cmkBADCAPT )
                        effort = cmkBADCAPT - 1;
                    effort |= (cmkBADCAPT >> 1);
//                    if( - new_hash_entry->val < -valMATEscore )
//                        effort = 0; // look that last

/*                    if( (depth & 1) == 0 ) {
                        effort = cmkBADCAPT - effort - 1;
                    }*/
                    pcm->cmk = mask | effort;
                }
                bUndoMove(pcon, pste, pcm->m);
            }
        }
    }

#endif
    //
    //	Iterate through the move list, trying everything out.
    //
    foreachmove(pste,pcm) {
        //
        //	All moves are sorted on the cmk field
        //	cmk is set in PrepareSort() and in other sort functions

        SortNext(pste, pcm);
        //
        //	Make the move

        bMove(pcon, pste, pcm->m);

        pste->danger = forced;

        stats.c_int++;
	    //int posVal = ValEval(pcon, pste, a, b);
        int extension = preExtend;
        // Note : now I allow the reduction after the extension,
        // this is needed for the check extension
        bool can_reduce = (extension) > 0 ? false : true;
        can_reduce = can_reduce && !forced;

        if( (pste + 1)->checkf ) {
            can_reduce = false;
            extension_check(pcon, pste, pcm, extension, a, depth);
        }
        extension_pawn(pcon, pste, phashp, pcm, extension, can_reduce, a, /*isPV*/true);

#ifdef RECAPTURE_EXTENSION
        if( Capt(pcm->m) && (To(pcm->m) == prevCaptureSqr)
//            && moveisGoodCapture(pcm)
            && moveisVeryGoodCapture(pcm)
//            && (a < valKNIGHT )
            ) {
            if( true ) {
                stats.recapture++;
                extension = 1;
                can_reduce = false;
            }
        }
#endif
//        if( Capt(pcm->m) ) {
//            extension_castle(pcon, pste, pcm, extension, can_reduce, a, /*isPV*/ true, kingRing);
//        }

        //  ==========================================================================
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( extension && (extCount < maxExt) ) {
            (pste + 1)->plyRem++;
        }
//        (pste+1)->c_altMoves = 0;
        U64 thisMoveNodeCount = pcon->ss.nodes;
        //
		//	Recurse into normal search or quiescent search, as appropriate.
		//
#ifdef USE_LMR
        // Note : LMR done on all non pv nodes moves
        //
        // move order is [H] [good capt] [equ capt] [killers] others [bad capt]
        if( false && 
            (c_move >= 2)
//            && (depth >= 2)
//            && (bestVal >= a)
            && can_reduce
            && !iaminCheck
            && (pste->plyRem >= 2)
//            && (pste->plyRem >= 1)
            && (a > -valMATEscore)
            && !moveisPromotion(pcm)
            && isNodeUseless(pcon, pste, 0/*hashMoveNodeCount*/, pcm)
            ) {
            stats.LMRhint++;
            // search with a reduced depth
            stats.LMR++;
            pste->reduced = true;

		    (pste + 1)->plyRem --;
/*            val = -Search(pcon, pste + 1, - b, - a, nodeCUT);
            // reduced was perhaps cleared
            if( val > a ) {
        		(pste + 1)->plyRem ++;
                pste->reduced = false;
                // re-search at normal depth
                stats.LMRreSearch++;
                val = -Search(pcon, pste + 1, - b, - a, nodeCUT);
            }
            pste->reduced = false;*/
        }
#endif
        if (!fFound) {
            val = -SearchPV(pcon, pste + 1, - Beta, - a);
            stats.c_pv++;
        } else {
            stats.c_pv_succ++;
            val = -Search(pcon, pste + 1, - b, - a, nodeCUT); // cut nodes

            if( pste->reduced ) {
                pste->reduced = false;
                if( val > a ) {
        		    (pste + 1)->plyRem ++;
                    // re-search at normal depth
                    stats.LMRreSearch++;
                    val = -Search(pcon, pste + 1, - Beta, - a, nodeCUT);
                }
            }

            if( val > a /*&& val < Beta*/) { // re-search
//                (pste+1)->c_altMoves = 0;
                stats.reSearch++;
                val = -SearchPV(pcon, pste + 1, - Beta, - a); // new PV
            }
        }


    thisMoveNodeCount = pcon->ss.nodes - thisMoveNodeCount;
/*    {
        PHASH hash_entry = ProbeHashValue(pste+1);
        if( hash_entry ) {
            hash_entry->count += U32(thisMoveNodeCount);
        }
    }*/
/*    if( (depth <= 3) && (thisMoveNodeCount > 1) ) {
        char tmp[256];
        sprintf(tmp, "#> stats PV : d=%d #=%d v=%d n=%I64u ", pcon->ss.plyDepth, c_move, val, thisMoveNodeCount);
        VDumpLine(pcon, pste, tmp);
    }
*/

        bUndoMove(pcon, pste, pcm->m);
        if ((pcon->fAbort) || (pcon->fTimeout)) {
            return 0;
        }
        //	We got a value back.  We unmade the move.  We're not dead.  Let's
        //	see how good this move was.  If it was >= "beta", it was so good
        //	that we don't need to search for anything better, so we'll leave.
        //
        //	If it was not >= "beta", but it was > "alpha", this is better than
        //	anything else we've found before, but not so good we have to
        //	leave.  These kinds of moves are actually quite rare.  If I find
        //	one of these, I have to store it in the PV (main-line) that I'm
        //	constructing.  This might end up being the main-line for the whole
        //	search, if it gets backed up all the way to the root.
        //
        if( val > bestVal ) {
            bestVal = val;
            if (val > a) {
                cmBest = *pcm;
                copyPV( pste, cmBest );
/*                if( pcon->ss.plyDepth == 14 && abs(val) == 34 ) {
                    VDumpSearch(pcon, pste, val, a, Beta);
                }*/
                if (val >= Beta) {
                    //
                    //	This move failed high
                    //
                    goto cutNode;
                }
                //	This move is between alpha and beta, which is actually pretty
                //	rare.  If this happens I have to add a PV move, and append the
                //	returned PV to it
                //
                a = bestVal;
                rankCut = iMove+1;
            }
        }
        fFound = true;
        c_move++;
//        pste->ttmove = false;

#ifdef NULL_WINDOW
        b = a + 1;              // set the new null window
#endif
    }

    if (!fFound) {
        // We have no move !!
        if (iaminCheck) {
            // Checkmate
            bestVal = mateValue(depth);
        } else {
            // Draw
            bestVal = 0;
        }
        // ==================================
        return bestVal;
        // ==================================
    }
    // if I have a move, then this is not an all nodes
cutNode:;

    if( bestVal > Alpha ) {
        pste->lastMove = cmBest.m;
        // for nm refutation
        (pste+1)->lastMove = cmBest.cmk;

        if( cmBest.cmk & cmkHASH )
            stats.c_cut_H++;
        else if( cmBest.cmk & cmkKILLER )
            stats.c_cut_killer++;
        else if( moveisGoodCapture(&cmBest) )
            stats.c_cut_capt++;
        else if( cmBest.cmk & cmkBADCAPT )
            stats.c_cut_badcapt++;
        else
            stats.c_cut_quies++;

        if( bestVal >= Beta ) {
            stats.cCUT++;
            if(iMove == 0)
                stats.cCUTrank1++;
            if(iMove < 5)
                stats.cCUTrankn++;
        } else {
            if( depth == pcon->ss.plyDepth ) {
                stats.cCUTa += lmc;
                stats.cCUTaRank += rankCut;
            }
//            stats.c_altMoves += altMoves;
            stats.c_alpha_cut++;
        }

        History.update( &cmBest, depth );

#ifdef USE_KILLERS
        //  Set the new killer move only if it was not marked as good allready
        //  'Good' moves are allready sorted first so storing them in the killer
        //  array is counter productive
        if( bestVal >= valMATEscore )
            MateKillers[depth].set(&cmBest);
        else if( !moveisCapture(&cmBest) )
            Killers[depth].set(&cmBest);
#endif
    } else
        stats.c_all++;

    if( ! pcon->fTimeout ) {
        RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, false);
    }
    Assert(ValIsOk(bestVal));
	return bestVal;
}

//  -------------------------------------------------------------------------

//  Full-width search.  This is the heart of the "chess playing" part of this
//  program.  It performs a recursive alpha-beta search.

static struct {
    char    move[12];
    int     c_nodes;
    int     hval;
} sortDebug[200];
static int c_sortDebug;

#define maxRootMoves    200
static struct _rootMove {
    char szmove[8];
    int score;
    int hint;
    U64 nodes;
    bool    evaluated;
    CM cm;
} rootMoves[maxRootMoves];
static int rootMovesCount;

static int sortScore( struct _rootMove & move) {
    return move.hint * 10000 + move.score;
}

int ValSearchRoot(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    int a = Alpha;
    int b = Beta;
    int bestVal = valMIN;
    int	val;
    int depth = 1;
    pste->lastMove = 0;
    pste->reduced = false;
    pste->c_ttpair = 0;
    pste->threatMove = U32(-1);
    pste->threatTarget = 0;
    pste->prevCaptSqr = -1;
    pste->danger = false;
    pste->fNull = false;

    c_sortDebug = 0;

    //	Increment global node counter
    //
    ++pcon->ss.nodes;

    //	Stuff is doone a little differently at the root than at any other
    //	ply.
    //
    //
    //	Mark this move so if I find it in a sub-tree I can detect a rep
    //	draw.  I don't check to see if a rep is already set, because if
    //	we are here, we have to be making a move, not whining about a
    //	blown chance for a draw.  I also don't try to cut off from the
    //	hash table, for the same reason.  I want a move, not a score.
    //
    //	I do probe the table though, in case there is "best move" info
    //	in there.
    //
    PHASH hash_entry = ProbeHashValue(pste);
    //
    //	If it's the first time we come here (depth == 0)
    //	generate the moves
    //
    if( pste->plyRem == 0 ) {
        // first time with this position
        bGenMoves(pcon, pste);
        pcon->ss.ccmLegalMoves = legalMoveCount(pste);
        rootMovesCount = pcon->ss.ccmLegalMoves;

        PHASHPawn phashp;
        EvalPawns(pcon, pste, phashp);
        PrepareSort( pcon, pste, depth, phashp );

        foreachmove(pste,pcm) {
            CmToSz( pcm, rootMoves[iMove].szmove );
            rootMoves[iMove].cm     = *pcm;
            rootMoves[iMove].hint   = 0;
            rootMoves[iMove].evaluated = false;

            bMove(pcon, pste, pcm->m);
            (pste+1)->plyRem = -2;  // -2 no check, no nm search
            int rootScore = -ValEval(pcon, pste+1, valMIN, valMAX);
//            int val = -SearchQ(pcon, pste + 1, -rootScore-500, -rootScore+500+1, true);
//            int val = -rootScore;
            int val = -SearchQ(pcon, pste + 1, -rootScore, -rootScore+1, true);
            bUndoMove(pcon, pste, pcm->m);
            rootMoves[iMove].score  = val; // keep the generation order

            rootMoves[iMove].nodes  = 0;
        }
    }
    {
        // after the first time (first depth iteration)
        // sort the moves by node count
#ifdef USE_HASH
        PHASH phash = hash_entry;
        U32 hm = phash ? phash->getMove() : 0;
#endif
        for(int i = 0; i < rootMovesCount; i++) {
            PCM pcm = &rootMoves[i].cm;
            if ( rootMoves[i].cm.m == lastPV[1].m )
                rootMoves[i].hint = 10;
#ifdef USE_HASH
		    else if ( pcm->m == hm )
                rootMoves[i].hint = 1;
#endif
        }
    }
	//	Mark the PV and hash moves so they sort first.
	//
    // bubble sort
    for(int i = 0; i < rootMovesCount - 1; i++) {

        for(int j = i+1 ; j < rootMovesCount; j++)
            // most moves have the same score because of null window search
            if( (sortScore(rootMoves[j]) > sortScore(rootMoves[i]) ||
                (sortScore(rootMoves[j]) == sortScore(rootMoves[i])) && (rootMoves[j].nodes > rootMoves[i].nodes) )) {
                _rootMove swap = rootMoves[i];
                rootMoves[i] = rootMoves[j];
                rootMoves[j] = swap;

            }
    }
    PHASHPawn phashp;
    EvalPawns(pcon, pste, phashp);
    bool forced = false;
    int preExtend = 0;
    bitboard kingRing = 0;

    before_search(pcon, pste, /*isPV*/ true, kingRing, forced, preExtend, 0, depth);

    CM cmBest(0,0);
	bool fFound = false;
	//
	//	Iterate through the move list, trying everything out.
	//
    for(int i = 0; i < rootMovesCount; i++) {
        PCM pcm = &rootMoves[i].cm;
//        pcm->cmk = i;

        //	Record information about which move I am searching ...
        //
        SAN	san;
        pcon->ss.icmSearching = i+1;
        CmToSan(pcon, pste, pcm, &san);
        SanToSz(&san, pcon->ss.aszSearching);
        CmToSz(pcm, pcon->ss.aszSearchSan);
        rootMoves[i].nodes = pcon->ss.nodes;
        VPrSendUciCurrMove( pcon->ss );

        //
        //	... then go make the move.
        //
        bMove(pcon, pste, pcm->m);
        bool can_reduce = true;
        int extension = 0;
	    (void) ValEval(pcon, pste, a, b);

        if( (pste + 1)->checkf ) {
            can_reduce = false;
            extension_check(pcon, pste, pcm, extension, a, depth);
        }
        extension_pawn(pcon, pste, phashp, pcm, extension, can_reduce, a, /*isPV*/true);
//        if( Capt(pcm->m) ) {
//            extension_castle(pcon, pste, pcm, extension, can_reduce, a, /*isPV*/ true, kingRing);
//        }

        //
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1 + extension;

		//
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(fFound)
            succNType = nodeCUT;
        else
            succNType = nodePV;

        if (!fFound) {
            val = -SearchPV(pcon, pste + 1, - Beta, - a);
        } else {
            val = -Search(pcon, pste + 1, - b, - a, succNType); // cut nodes
//            val = -SearchPV(pcon, pste + 1, - (a+1), - a); // cut nodes
            if( val > a && !((pcon->fAbort) || (pcon->fTimeout))) { // re-search
                stats.reSearch++;
                val = -SearchPV(pcon, pste + 1, - Beta, - a); // new PV
            }
        }

        // update the next sort key
        rootMoves[i].hint  = 0;
        if( ! rootMoves[i].evaluated )
            rootMoves[i].score = val;
        rootMoves[i].nodes = pcon->ss.nodes - rootMoves[i].nodes;

        // debugging stuff
        sortDebug[c_sortDebug].c_nodes = int(pcon->ss.nodes) - sortDebug[c_sortDebug].c_nodes;
        c_sortDebug++;

		bUndoMove(pcon, pste, pcm->m);
		if ((pcon->fAbort) || (pcon->fTimeout)) {
			return 0;
		}
		//	We got a value back.  We unmade the move.  We're not dead.  Let's
		//	see how good this move was.  If it was >= "beta", it was so good
		//	that we don't need to search for anything better, so we'll leave.
		//
		//	If it was not >= "beta", but it was > "alpha", this is better than
		//	anything else we've found before, but not so good we have to
		//	leave.  These kinds of moves are actually quite rare.  If I find
		//	one of these, I have to store it in the PV (main-line) that I'm
		//	constructing.  This might end up being the main-line for the whole
		//	search, if it gets backed up all the way to the root.
        //
        if( val > bestVal ) {
            bestVal = val;
            if (val > a) {
                if (val >= Beta) {
                    // Note : not reached if not doing aspiration search
                    //
                    //	This move failed high, so we are going to return beta,
                    //	but if we're sitting at the root of the search I will set
                    //	up a one-move PV (if this isn't already the move I'm
                    //	planning to make), so this move will be made if I run out
                    //	of time before resolving the fail-high.
                    //
                    copyPV( pste, *pcm );

                    //	This function needs the root moves generated, which
                    //	they are.
                    //
                    RecordHashValue(pcon, pste, *pcm, bestVal, Alpha, Beta, depth, false);
                    Beta = val;
                    VDumpPv(pcon, pcon->ss.plyDepth, Beta, prsaFAIL_HIGH);
                    return bestVal;
			    }
                //	This move is between alpha and beta, which is actually pretty
                //	rare.  If this happens I have to add a PV move, and append the
                //	returned PV to it, and if I'm at the root I'll send the PV to
                //	the interface so it can display it.
                //
                rootMoves[i].score = val;
                rootMoves[i].evaluated = true;
                cmBest = *pcm;
                copyPV( pste, cmBest );
                a = bestVal;
                //
                //	This function needs the root moves generated, and they
                //	are.
                //
                // clear failed high, can be set again in VDumpPV is val is bad
                pcon->ss.failed = false;
                VDumpPv(pcon, pcon->ss.plyDepth, a, prsaNORMAL);
#if 1 //_DEBUG
                if( TmNow() - pcon->ss.tmStart > 1000 )
                    stats.report();
#endif
            }
        }
#ifdef NULL_WINDOW
        b = a + 1;              // null window
#endif
        fFound = true;
    }
#if 0
    printf("# SORT was ");
    for(int i = 0; i < c_sortDebug && i < 10; i++) {
        if( sortDebug[i].c_nodes < 10000 )
            printf("%s (%d) ", sortDebug[i].move, sortDebug[i].c_nodes );
        else
            printf("%s (%dk) ", sortDebug[i].move, sortDebug[i].c_nodes / 1000 );
        if( sortDebug[i].hval > 0 )
            printf("%d ", sortDebug[i].hval );
    }
    printf("\n");
#endif
    if (!fFound) {
        // We have no move !!
        if (pste->checkf) {
            // Checkmate
            bestVal = mateValue(depth);
        } else {
            // Draw
            bestVal = 0;
        }
    }
    for(int i = 0; i < rootMovesCount; i++) {
        // only use node count
        rootMoves[i].score = 0;
        rootMoves[i].evaluated = false;
    }

    if( ! pcon->fTimeout )
        RecordHashValue(pcon, pste, cmBest, bestVal, Alpha, Beta, depth, false);
    Assert(ValIsOk(bestVal));
    return bestVal;
}
